
#include <fstream.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <graphics.h>
#include <math.h>

class Data
{
    private:
        int vertex;
        float distance;
    public:
        void Set_vertex(int v)
        {
            vertex=v;
        }
        int Return_vertex()
        {
            return vertex;
        }

        void Set_distance(float d)
        {
            distance=d;
        }
        float Return_distance()
        {
            return distance;
        }

};


//__________Nodes for the heap structure_______________
class Node
{
    private:
        Data data;
    public:

        void Set_priority(float p)
        {
            data.Set_distance(p);
        }
        float Return_priority()
        {
            return data.Return_distance();
        }

        void Set_data(Data d)
        {
            data=d;
        }
        Data Return_data()
        {
            return data;
        }


};

class Heap
{
    private:
        Node* heap;
        int last;		 //____last stores current last element position_________
        int size;
        int* node_locations; //_____stores positions of nodes in the heap______

    public:
        //____________Constructors____________
        Heap(int siz)
        {
            heap=new Node[siz];
            node_locations=new int[siz];
            int i;
            for(i=0;i<siz;i++)
            {
                node_locations[i]=-1;
            }

            last=-1;
            size=siz;
        }

        ~Heap()
        {
            Destroy();
        }

        void Destroy()
        {
            delete[] heap;
            heap=NULL;
            delete[] node_locations;
            node_locations=NULL;
        }

        //___________Interface Functions_________
        int Insert_Element(Data d);
        Data DeleteMinimum();
        int Update_Element(Data d);
        Data Return_data(int vertexno)
        {
            int pos=node_locations[vertexno-1];
            if(pos>=0)
            {
                return heap[pos].Return_data();
            }
            else
            {
                Data d;
                d.Set_vertex(-1);
                return d;
            }
        }


        void ClearHeap()
        {
            last=-1;
        };
        int IsEmpty()
        {
            if (last<0)
                return -1;
            else
                return 0;
        }

        void Print()
        {
            int i;
            for(i=0;i<=last;i++)
            {
                cout << endl << heap[i].Return_data().Return_vertex();
                cout << "    " << heap[i].Return_data().Return_distance();
            }
            for(i=0;i<=last;i++)
            {
                cout << endl << node_locations[i];
            }
        }


    private:
        //_________Implementation Functions______________
        void SwapNodes(Node& a,Node& b)
        {
            Node t;
            t=a;
            a=b;
            b=t;
        }
        void SwapNodeLocations(int a,int b)
        {
            int t;
            t=node_locations[a];
            node_locations[a]=node_locations[b];
            node_locations[b]=t;
        }

        int Parent(int c)
        {
            if(c==0)
                return 0;
            else
                return int((c+1)/2  -  1);
        }

        int LeftChild(int p)
        {
            return 2*p+1;
        }

        int RightChild(int p)
        {
            return (2*p+2);
        }


};


int Heap::Insert_Element(Data d)
{

    if(last==size-1)
        return -1;
    else
    {
        last++;
        heap[last].Set_data(d);
        node_locations[d.Return_vertex()-1]=last;
    }

    int i=last;
    //_______Swap up till we get heap property_________
    while( heap[Parent(i)].Return_priority() >
            heap[i].Return_priority() )
    {

        SwapNodes(heap[Parent(i)],heap[i]);

        SwapNodeLocations(heap[Parent(i)].Return_data().Return_vertex()-1,
                          heap[i].Return_data().Return_vertex()-1);

        i=Parent(i);
    }
    return 0;


}

int Heap::Update_Element(Data d)
{
    //	cout << endl << "Update " << d.Return_vertex() << " "  << d.Return_distance();
    //	getch();

    if(node_locations[d.Return_vertex()-1]==-1)
    {
        return -1;
    }
    else
    {
        heap[node_locations[d.Return_vertex()-1]].Set_data(d);
    }

    int i=node_locations[d.Return_vertex()-1];
    //_______Swap up till we get heap property_________
    while( heap[Parent(i)].Return_priority() >
            heap[i].Return_priority() )
    {

        SwapNodes(heap[Parent(i)],heap[i]);

        SwapNodeLocations(heap[Parent(i)].Return_data().Return_vertex()-1,
                          heap[i].Return_data().Return_vertex()-1);

        i=Parent(i);
    }
    return 0;


}

Data Heap::DeleteMinimum()
{
    if(last<0)
    {
        cout << "Error : No Element Left !";
        return *(new Data);
    }

    //_________Store minimum priority element__________
    Data rt=heap[0].Return_data();
    node_locations[rt.Return_vertex()-1]=-1; //___Indicate not in Heap_____

    //_________Assign last element to first__________
    heap[0]=heap[last];
    node_locations[heap[last].Return_data().Return_vertex()-1]=0;
    last--;

    int i=0;
    //__________Keep on swapping downwards with lower priority child till
    //__________penultimate level of tree________________________
    while( i<=Parent(last) && last>0)
    {
        int j;

        if( ( heap[LeftChild(i)].Return_priority() <
                heap[RightChild(i)].Return_priority() ) || LeftChild(i)==last)
        {
            j=LeftChild(i);
        }
        else
            j=RightChild(i);

        if(heap[i].Return_priority() > heap[j].Return_priority() )
        {
            SwapNodes(heap[i],heap[j]);
            SwapNodeLocations(heap[i].Return_data().Return_vertex()-1,
                              heap[j].Return_data().Return_vertex()-1);

        }

        i=j;
    }

    return rt;
}


class LinkList
{
    public:
        int vertex;
        float distance;
        LinkList* next;

    public:
        //_______Constructor______
        LinkList(int v,float d,LinkList* n)
        {
            vertex=v;
            distance=d;
            next=n;

        }

        //_______Interface functions__________
        LinkList*  Insert_Element(int v,float d)  //_______Adds in sorted order________
        {
            if(v>vertex)
            {
                if(next!=NULL)
                {
                    next=next->Insert_Element(v,d);
                }
                else
                {
                    next=new LinkList(v,d,NULL);
                }
                return this;
            }
            else
                return new LinkList(v,d,this);

        }

        //________Useful for traversals____________
        LinkList* Return_next()
        {
            return next;
        }

        int Return_vertex()
        {
            return vertex;
        }

        float Return_distance()
        {
            return distance;

        }

};

//___________Stores the graph__________
class GraphMap
{
    private:
        LinkList** array;
        int size;
        int noofedges;

    public:
        //_______Constructors__________
        GraphMap(int no_of_vertex) //______Please specify size_______
        {
            array=new LinkList*[no_of_vertex];
            size=no_of_vertex;
            //______Initialize to Null______
            for(int i=0;i<no_of_vertex;i++)
            {
                array[i]=NULL;
            }

            noofedges=0;
        }

        void Destroy()
        {
            LinkList* travel;
            for(int i=0;i<size;i++)
            {
                travel=array[i];
                while(travel!=NULL)
                {
                    LinkList* t=travel->Return_next();
                    if(travel)
                        delete travel;
                    travel=t;
                }
                array[i]=NULL;
            }
            delete array;
            array=NULL;
        }

        void Clear()
        {
            LinkList* travel;
            for(int i=0;i<size;i++)
            {
                travel=array[i];
                while(travel!=NULL)
                {
                    LinkList* t=travel->Return_next();
                    if(travel)
                        delete travel;
                    travel=t;
                }
                array[i]=NULL;
            }

            noofedges=0;
        }


        void Insert_Element(int startv,int endv,float distance)
        {
            if(array[startv-1]==NULL)
            {
                array[startv-1]=new LinkList(endv,distance,NULL);
            }
            else
            {
                array[startv-1]=array[startv-1]->Insert_Element(endv,distance);
            }
            noofedges++;
        }

        int Return_noofedges()
        {
            return noofedges;
        }

        void Set_noofedges(int e)
        {
            noofedges=e;
        }

        LinkList* Return_LinkListPointer(int v)
        {
            return array[v-1];
        }

        int EdgeIsPresent(int sv,int ev,float d)
        {
            LinkList* travel;
            travel=Return_LinkListPointer(sv);
            while(travel!=NULL)
            {
                if(travel->Return_vertex()==ev && travel->Return_distance()==d)
                {
                    return 1;
                }
                travel=travel->Return_next();
            }
            return 0;
        }

        int EdgeIsPresent(int sv,int ev)
        {
            LinkList* travel;
            travel=Return_LinkListPointer(sv);
            while(travel!=NULL)
            {
                if(travel->Return_vertex()==ev)
                    return 1;
                travel=travel->Return_next();
            }
            return 0;
        }

        void Print()
        {
            LinkList* travel;
            for(int i=0;i<size;i++)
            {
                travel=Return_LinkListPointer(i+1);

                cout << endl << (i+1) << " -  ";
                while(travel!=NULL)
                {
                    cout << travel->Return_vertex() << "," << travel->Return_distance() << "-";
                    travel=travel->Return_next();
                }

            }
        }


};


class Diktra
{
    private:
        Heap* heap;
        LinkList* s;
        GraphMap* g;
        int source_vertex;
        int no_of_vertices;
        int  error;

    public:
        Diktra(char* filename)
        {

            source_vertex=1;
            heap=NULL;
            s=NULL;
            g=NULL;


            error=0;
            error=FillGraph(filename);// ________Init Graphmap_______

            //		g->Print();
            //		getch();

            InitHeap(); //_____All all vertices with distance infinity______

            s=NULL;
            //_______Update source in heap________

        }

        ~Diktra()
        {
            Destroy();
        }

        void Destroy()
        {
            if(heap)
                heap->Destroy();
            delete heap;
            heap=NULL;

            if(g)
                g->Destroy();
            delete g;
            g=NULL;
        }


        //_________Interface_________
        void Run();

        LinkList* Return_s()
        {
            return s;
        }
        GraphMap* Return_g()
        {
            return g;
        }
        void Set_source_vertex(int sv)
        {
            source_vertex=sv;
            Data d;
            d.Set_vertex(sv);
            d.Set_distance(0);
            heap->Update_Element(d);
        }

        int Return_no_of_vertices()
        {
            return no_of_vertices;
        }

        int Return_error()
        {
            return error;
        }





    private:
        int FillGraph(char* fname);
        void InitHeap();

};

void Diktra::Run()
{
    Data source,d;
    LinkList* travel;


    while(!heap->IsEmpty())
    {
        source=heap->DeleteMinimum(); //_______Get Least Element______

        //			cout << endl << "Min : " << source.Return_vertex() << "   " << source.Return_distance();
        //			getch();

        //______Add it in set 's'___________
        if(s==NULL)
        {
            s=new LinkList(source.Return_vertex(),source.Return_distance(),NULL);
        }
        else
        {
            s=s->Insert_Element(source.Return_vertex(),source.Return_distance());
        }


        travel=g->Return_LinkListPointer(source.Return_vertex());

        //________Now Update all distances_________
        while(travel!=NULL)
        {
            //				cout << endl << "Travelling  : " << travel->Return_vertex();
            //				getch();

            d=heap->Return_data(travel->Return_vertex());

            //				cout << endl << "Data returned : " << d.Return_vertex() << " " << d.Return_distance();
            //				getch();

            if(d.Return_vertex()>0)
            {
                float newd=source.Return_distance()+travel->Return_distance();

                if(newd<d.Return_distance() && newd>=0)
                {
                    d.Set_distance(newd);
                    heap->Update_Element(d);
                }
            }
            travel=travel->Return_next();
        }
    }
}


void Diktra::InitHeap()
{
    Data d;
    int i;

    for(i=1;i<=no_of_vertices;i++)
    {
        d.Set_vertex(i);
        d.Set_distance(INF);
        heap->Insert_Element(d);
    }
}


int Diktra::FillGraph(char* fname)
{
    ifstream infile;
    infile.open(fname);
    if(infile==0)
    {
        return -1;
    }



    int i,j,l;
    int noofvertex;
    int edgeno;

    infile >> noofvertex;
    if(source_vertex>noofvertex)
        return (-2);
    no_of_vertices=noofvertex;

    infile >> edgeno;

    //______init heap and graphmap_______
    heap=new Heap(noofvertex);
    g=new GraphMap(noofvertex);


    int sv,ev;
    float distance;
    for(i=1;i<=edgeno;i++)
    {
        infile >> sv >> ev >> distance;
        sv++;
        ev++;
        if(distance>INF)
            distance=INF; //________constrain distances > INF
        if(distance<0)
            distance=0;
        g->Insert_Element(sv,ev,distance);
    }


    infile.close();

    return 0;
}

