/*$T simpeer.cpp GC 1.137 03/07/03 12:13:39 */


/*$6
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 TODO
 
 
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */


#include <fstream.h>
#include <time.h>
#include <stdio.h>
#include <vector>

#define MAXEVENTS	10000
#define MAX_FRIENDS 500
#define DEBUG		2

enum EVENTTYPE { RECEIVE, SEND };

class INFORMATION
{
    public:

        int		length;
        int ip;
        char	information[256];
};

/*
 =======================================================================================================================
    this has 6 fields
 =======================================================================================================================
 */
class	EventData
{

    public:
        float			time;
        EVENTTYPE		type;
        unsigned long	to;
        float			freshness;
        unsigned long	from;
        INFORMATION		information;

        void print()
        {
            cout << "Event : time=" << time << ", type = " << type << ", to = " << to << ", from = " << from << ", Information = " << information.information << endl;
        }

};

class	EventManager
{
    public:
        EventData	eventlist[MAXEVENTS];

        EventManager(void)
        {
            init();
        }

        void init(void)
        {
            for(int i = 0; i < MAXEVENTS; i++)
                eventlist[i].time = -1;
        }

        void add_event(EventData event)
        {
            /* Add event */
            for(int i = 0; i < MAXEVENTS; i++)
            {
                if(eventlist[i].time < 0)
                {
                    eventlist[i] = event;
                    return;
                }
            }
        }

        EventData get_min_event(void)
        {
            /*~~~~~~~~~~~~~~~~~~~~~*/
            /* Return minimum Time related with the ID */
            float	min_time = 9999999999;
            int		min_id;
            /*~~~~~~~~~~~~~~~~~~~~~*/

            for(int i = 0; i < MAXEVENTS; i++)
            {
                if(eventlist[i].time < min_time && eventlist[i].time>0)
                {
                    min_time = eventlist[i].time;
                    min_id = i;
                }
            }

            EventData min_event=eventlist[min_id];
            eventlist[min_id].time=-1;
            return min_event;
        }
};

class	Global
{

    public:
        float			current_time;
        EventManager	event_manager;


        Global()
        {
            current_time=0;  //creation time
        }

        float decay_freshness(float freshness,float latency)
        {

            return freshness-latency/50.0; // linear model of decay

        }

        float get_latency(unsigned long from, unsigned long to)
        {
            return 8;
        }
}
global;


class	DATABASE
{
    public:
        unsigned long		ip;		/* this if friend id here now in the simulator */
        float				latency;
        float				freshness;
        class INFORMATION	information;
        char				valid;	/* 0 if invalid, 1 if valid */

        void print()
        {
            cout << "DATABASE: ip=" << ip << ",latency=" << latency << ",freshness="<<freshness << ",information=" <<
            information.information << ",valid=" << (int)valid << endl;

        }
};


class	Peer
{
    public:
        unsigned long					myid;

        vector<DATABASE>					database_vector;
        vector<int>							latency_vector;
        vector<int>							freshness_vector;
        vector<vector<DATABASE>::iterator>	send_vector;

        Peer(unsigned long id)
        {
            if(DEBUG>=2)
            {
                cout << "Entering constructor of Peer : " << id << endl;
            }
            myid = id;
            database_vector.reserve(MAX_FRIENDS);
            latency_vector.reserve(MAX_FRIENDS);
            freshness_vector.reserve(MAX_FRIENDS);


            //add_friend(myid); // add myself to the database list
            DATABASE d;
            d.ip=myid;
            d.latency=0;
            d.freshness=100;
            database_add(d);

            update_myinfo();

            // add an send event for myself
            EventData event;
            event.time = global.current_time + 0.5;
            event.to = myid;
            event.type = SEND;
            event.from = myid;
            global.event_manager.add_event(event);

            if(DEBUG>=2)
                cout << "Peer: " << myid << " Scheduling a send event for myself at time " << event.time << endl;

        }


        void print_vectors()
        {
            vector<DATABASE>::iterator	p;
            vector<int>::iterator lp;

            cout << "Printing the database vector....\n";
            /* check if friend already exists */
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                p->print();
            }

            cout << "Printing the latency vector....\n";
            for(lp = latency_vector.begin(); lp != latency_vector.end(); lp++)
            {
                database_vector[(*lp)].print();
            }

            cout << "Printing the freshness vector....\n";
            for(lp = freshness_vector.begin(); lp != freshness_vector.end(); lp++)
            {
                database_vector[(*lp)].print();
            }
        }


        void set_friend_list(void)
        {
            vector<DATABASE>::iterator	  p;

            send_vector.clear(); /* MULTICAST model , add all the people in the database */
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                if(p->ip != myid)
                    send_vector.push_back(p);
            }
        }

        int add_friend(unsigned long ip)
        {
            /*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
            vector<DATABASE>::iterator	p;
            /*~~~~~~~~~~~~~~~~~~~~~~~~~~*/

            if(DEBUG >= 2)
                cout << "Entering add_friend()\n";

            /* check if friend already exists */
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                if(p->valid && p->ip == ip)
                    return -1;        // friend already exists
            }

            /*~~~~~~~~~~*/
            DATABASE	d;
            /*~~~~~~~~~~*/

            d.ip = ip;
            d.valid = 0;
            d.latency = global.get_latency(myid, ip);
            d.freshness = -1;
            strcpy(d.information.information,"Friend !");

            database_add(d);

            if(DEBUG >= 2)
                cout << "Leaving the add_friend()\n";
            return 0;
        }



        /*
         ===================================================================================================================
            adds an element to the current database and also updates the latency and freshness pointer arrays
         ===================================================================================================================
         */
        void database_add(DATABASE d)
        {
            bool flag=false;
            database_vector.push_back(d);

            /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
            vector<DATABASE>::iterator						p;
            vector<int>::iterator	lp;
            /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

            p = database_vector.end();

            /* add to latency vector */
            for(lp = latency_vector.begin(); lp != latency_vector.end(); lp++)
            {
                if(database_vector[(*lp)].latency > d.latency)
                {
                    latency_vector.insert(lp, p-database_vector.begin()-1)
                    ;
                    flag=true;
                    break;
                }
            }

            if(flag==false)
                latency_vector.push_back(p-database_vector.begin()-1);

            flag=false;

            /* add to freshness vector */
            for(lp = freshness_vector.begin(); lp != freshness_vector.end(); lp++)
            {
                if(database_vector[(*lp)].freshness < d.freshness)
                {
                    freshness_vector.insert(lp,p-database_vector.begin()-1);
                    flag=true;
                    break;
                }
            }

            if(flag==false)
                freshness_vector.push_back(p-database_vector.begin()-1);
        }

        // this is to delete an element in the database
        void database_del(unsigned ip)
        {
            vector<DATABASE>::iterator	p;
            vector<int>::iterator lp;
            bool flag=false;

            cout << "Peer:" << myid << ", Entering database_del for ip :" << ip << endl;
            /* check if friend already exists */
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                if(p->ip==ip)
                {
                    flag=true;
                    break;
                }
            }

            for(lp = latency_vector.begin(); lp != latency_vector.end(); lp++)
            {
                if(database_vector[(*lp)].ip==ip)
                {
                    latency_vector.erase(lp);
                    break;
                }
            }

            for(lp = freshness_vector.begin(); lp != freshness_vector.end(); lp++)
            {
                if(database_vector[(*lp)].ip==ip)
                {
                    freshness_vector.erase(lp);
                    break;
                }
            }

            if(flag)
            {
                database_vector.erase(p);
                cout << "Delete succesful !\n";
            }
            else
                cout << "Failed to delete for ip : " << ip << endl;
        }

        void update_database(DATABASE d)
        {
            database_del(d.ip);
            database_add(d);

        }

        // update the database with updated information
        // requires updating the latency_vector and freshness_vector also, which is tricky

        void update_myinfo()
        {
            time_t t;
            t=time(NULL);
            sprintf(database_vector[0].information.information,"This is me, Mr. %u,Time is %s",myid,ctime(&t));

            database_vector[0].valid=1;
            database_vector[0].information.length=strlen(database_vector[0].information.information)+1;
            database_vector[0].information.ip=myid;


        }
        /*
         ===================================================================================================================
            schedule recieve information events for desired number of friends
         ===================================================================================================================
         */
        void event_send(EventData ev)
        {
            /*~~~~~~~~~~~~~~*/
            int			i, j,to;
            EventData	event;
            /*~~~~~~~~~~~~~~*/

            if(DEBUG>=2)
                cout << "Entering event_send() , myid = " << myid << endl;

            set_friend_list();
            if(send_vector.size()==0)
                cout << "Peer: " << myid << ", My send vector is empty !" << endl;
            else
                cout << "Peer: "  << myid << " Send_vector size is " << send_vector.size() << endl;

            for(i = 0; i < send_vector.size(); i++)
            {
                to = send_vector[i]->ip;

                for(j = 0; j < database_vector.size(); j++)
                {
                    if(database_vector[j].valid==0)
                        continue;
                    /* schedule a recieve event */
                    event.time = global.current_time + global.get_latency(myid, to);
                    event.to = to;
                    event.type = RECEIVE;
                    event.from = myid;
                    event.freshness = database_vector[j].freshness;
                    event.information = database_vector[j].information;
                    event.information.ip=database_vector[j].ip;    // to whom this information belongs

                    global.event_manager.add_event(event);
                }


                cout << "Peer : " << myid << ", scheduled a recieve event for " << to << " at time " << event.time << endl;


            }

            /* schedule a send event for myself */


            event.time = global.current_time + 5;
            event.to = myid;
            event.type = SEND;
            event.from = myid;
            strcpy(event.information.information,"Don't Care!");

            global.event_manager.add_event(event);
            if(DEBUG>=2)
                cout << "Peer: " << myid << " Scheduling a send event for myself at time " << event.time << endl;

        }

        /*
         ===================================================================================================================
            got a packet from a peer !
         ===================================================================================================================
         */
        void event_recieve(EventData event)
        {
            /* just add this packet to the database, that's all; */

            if(DEBUG >= 2)
                cout << "Entering event_recieve, myid = " << myid << endl;

            /* update the freshness */
            event.freshness = global.decay_freshness(event.freshness, global.get_latency(event.from, event.to));

            /* ideally also look at originating timestamp */
            DATABASE d;

            d.ip=event.information.ip;
            d.latency=global.get_latency(myid,d.ip);
            d.freshness=event.freshness;
            d.information=event.information;
            d.valid=1;

            update_database(d);
            //update the database
            cout << "Peer : " << myid << " , Printing my database now...\n";
            print_vectors();

            // till now we don't really call the add_friend function.

        }
};


void main(void)
{

    //declare 3 peers for now
    Peer* peer[3];

    //set the global time thing


    // init the peers
    cout << "Initializing the peers...\n";

    int i;
    for(i=0;i<3;i++)
    {
        peer[i]=new Peer(i);
        if(i>0)
            peer[i]->add_friend(i-1);
    }


    cout << "Now starting the event_handler loop\n";
    // start the event handler from now on
    EventData event;
    for(i=0;i<100;i++)
    {
        event=global.event_manager.get_min_event();


        global.current_time=event.time;

        cout << "\n********Step : " << i << " , Extracted event at time " << global.current_time << endl;
        event.print();


        switch(event.type)
        {
            case RECEIVE:
                peer[event.to]->event_recieve(event);
                break;

            case SEND:
                peer[event.to]->event_send(event);
                break;

            default:
                break;
        }
    }
}
