/*$6
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 TODO
 
 FIXED
 
 1. Fixed bugs in what and whom. Make sure that I have what items to send and whom friends to send
 otherwise take the minimum.
 2. Don't send any information to myself. ( previous bug)
 
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */


#include <fstream.h>
#include <iomanip.h>
#include <time.h>
#include <stdio.h>
#include <vector>
#include <queue>
#include <algorithm>
#include <sys/timeb.h>
#include <math.h>

using namespace std;


#define MAX_FRIENDS 1024
#define MAX_PEERS 1024
#define DEBUG		0
#define PI 3.1415
#define PERPEER 0
#define MAXV 1024
#define INF 9999999999.0

enum EVENTTYPE { RECEIVE, SEND, PROBE };


class INFORMATION
{
    public:

        int		length;
        int ip;
        //char	information[256];
};

class edge
{
    public:
        int v;
        double latency;

        edge(int a, double b)
        {
            v=a;
            latency=b;
        }
};



class Topology
{
    public:

        vector<edge> graph[MAXV];

        vector<int> parentv; // stores the tree after the shortest path algorithm
        vector<double> result; // stores the shortest cost paths to all destinations

        double latency_matrix[1024][1024];

        int v,e;  // total number of vertices till now



        Topology()
        {
            v=0;
        }

        // reads the topology into our datastructure
        void read_topology(char* file)
        {
            ifstream f;
            f.open(file);
            if(f==0)
            {
                cout << "\nerror opening file : " << file << endl;
                return;

            }
            int n1,n2;
            double delay;
            int i;

            f >> v >> e;
            for(i=0;i<e;i++)
            {
                f >> n1 >> n2 >> delay;
                edge e(n2,delay);
                graph[n1].push_back(e);
                edge e1(n1,delay);
                graph[n2].push_back(e1);
            }
            f.close();

            cerr << "\nGenerated the topology...Printing it...";
            if(DEBUG>=2)
            {
                print_graph();
            }
            cerr <<"\nNow Generating latency latency_matrix...\n";

            make_latency_matrix();

            if(DEBUG>=2)
            {
                print_matrix();
            }
        }

        void print_matrix()
        {
            int i,j;
            cout << "\nPrinting the latency_matrix...." << endl;
            for(i=0;i<v;i++)
            {
                for(j=0;j<v;j++)
                {
                    cout << i<<","<<j<<":"<<latency_matrix[i][j] << " ; ";
                }
                cout << endl;
            }
        }

        // run the dijktra algorithm for each source node and
        // construct the latency latency_matrix
        void make_latency_matrix()
        {
            int i,j;
            for(i=0;i<v;i++)
            {
                dijktra(i);
                for(j=0;j<v;j++)
                {
                    latency_matrix[i][j]=result[j];
                }
            }
        }

        // for a given source, finds the shortest path to all the rest of the nodes
        void dijktra(int src)
        {
            //cout << "Entering djiktra()....";

            // v is the number of vertices in the graph
            vector<double> heap(v);
            vector<bool> extracted(v);
            result.reserve(v);
            parentv.reserve(v);


            int i;
            for(i=0;i<v;i++)
            {
                heap[i]=INF;


                parentv[i]=-1;
                result[i]=INF;
                extracted[i]=false;
            }

            heap[src]=0;
            parentv[src]=src;
            int ex=0;

            //cout << "starting while";


            while(ex<v)
            {
                vector<double>::const_iterator min=min_element(heap.begin(),heap.end());
                int v=min-heap.begin();

                //cout << "\nExtracted : " << v;
                cout.flush();

                extracted[v]=true;
                result[v]=heap[v];
                heap[v]=INF;

                ex++;
                unsigned int i;
                for(i=0;i<graph[v].size();i++)
                {
                    int w=graph[v][i].v;
                    if(extracted[w]==false && heap[w]>(graph[v][i].latency+result[v]))
                    {
                        heap[w]=graph[v][i].latency+result[v];
                        parentv[w]=v;
                    }
                }
            }
        }

        void print_graph() const
        {
            int i;
            cerr << "Entering print_graph(). V = " << v << endl;

            for(i=0;i<v;i++)
            {

                cerr << i << " : ";

                for(vector<edge>::const_iterator iterator= graph[i].begin();iterator!= graph[i].end();iterator++)
                {
                    cerr << iterator->v << "," << iterator->latency << " ; ";
                }
                cerr << endl;

            }

        }
};

/*
 =======================================================================================================================
    this has 6 fields
 =======================================================================================================================
 */
class	EventData
{

    public:
        float			time;
        EVENTTYPE		type;
        unsigned long	to;
        float			freshness; // derived from time information
        unsigned        heartbeat;
        unsigned long	from;
        INFORMATION		information;


        EventData()
        {
            // strcpy(information.information,"");
            freshness=0;
            from=to=0;
            heartbeat=0;
        }
        void print()
        {
            cout << "Event : time=" << time << ", type = " << type << ", from = " << from << "--->  to = " << to << " Heartbeat : " << heartbeat  << endl;
        }

};

struct SmallerThan
{
    bool operator()( EventData first,  EventData second) const
    {
        return (first.time > second.time)? 1 : 0;
    };
};

class	EventManager
{
    public:

        priority_queue<EventData,vector<EventData>,SmallerThan> event_list;
        EventData top;

        EventManager(void)
        {}



        void add_event(EventData event)
        {
            /* Add event */
            event_list.push(event);
            //cerr << "Adding event !, event_list size is now : " << event_list.size() << endl;


        }

        EventData get_min_event(void)
        {
            /*~~~~~~~~~~~~~~~~~~~~~*/
            /* Return minimum Time related with the ID */

            top=event_list.top();
            event_list.pop();
            //cerr << "Removing event !, event_list size is now : " << event_list.size() << endl;

            return top;
        }
};

char* s_SCHEMES[]={"BROADCAST",
                   "RANDOM",
                   "SPATIAL"};

enum SCHEMES {BROADCAST,RANDOM,SPATIAL};

class Peer;
class	Global
{

    public:
        float			current_time;
        EventManager	event_manager;
        unsigned        no_of_peers;
        Peer*           peer[MAX_PEERS];
        Topology topology;


        //global_metrics

        unsigned messages_sent;
        unsigned bytes_sent;
        double average_freshness;
        double worst_freshness;
        double average_database_size;
        double average_oracle_freshness;
        double count_of_fresh_items;
	unsigned heart_beat_distribution[100];

        double sigma; // for oracle freshness

        // gossip scheme stuff
        int what_scheme, whom_scheme;

        //--------------------------
        //  SCHEME PARAMETERS
        //------------------------

        // BROADCAST
        // no parameters here

        // RANDOM
        int whom;
        int what;

        // SPATIAL
        double ro;
        double cx;


        Global()
        {
            current_time=0;  //creation time
            average_oracle_freshness=0;
            sigma=20.0;
         }

        double decay_freshness(float freshness,float latency)
        {
             return freshness-latency/20.0; // linear model of decay

        }

        double oracle_freshness(unsigned heartbeat_difference)
        {
            //double result = exp(-double(heartbeat_difference*heartbeat_difference)/(2*sigma*sigma));

	    double result = heartbeat_difference;
            if(DEBUG>=3)
            {
                cout << "Oracle freshness () : heartbeat_difference : " << heartbeat_difference << "," <<   (sigma*(sqrt(2*PI))) << endl;
                cout << "oracle_freshness() : " << result << endl;
            }

            return result;
        }

        float get_latency(unsigned long from, unsigned long to)
        {
            return topology.latency_matrix[from][to];
        }
}
global;


class	DATABASE
{
    public:
        unsigned long		ip;		/* this if friend id here now in the simulator */
        double receipt_time; // to store when did we recieve this packet
        float				latency;
        float				freshness;
        unsigned            heartbeat;  // records the originating heartbeat of this item
        float oracle_freshness;    // metric computed to see the difference in heartbeat of this item
        class INFORMATION	information;
        char				valid;	/* 0 if invalid, 1 if valid */

        DATABASE()
        {
            //strcpy(information.information,"blank !");
            freshness=0;
            oracle_freshness=0;
            valid=0;
            heartbeat =0;
        }
        void print()
        {
            /*
               cout << "DATABASE: ip=" << ip << ",latency=" << latency << ",freshness="<<freshness << ",information=" <<
               information.information << ",valid=" << (int)valid << endl;*/

            cout << "DATABASE: ip=" << ip << ",latency=" << latency << ",freshness="<<freshness << ",valid=" << (int)valid << endl;
        }
};


inline unsigned min(unsigned a, unsigned b)
{
    if(a<b)
        return a;
    return b;
}

class	Peer
{
    public:
        unsigned long												myid;

        vector<DATABASE>									database_vector;
        vector<DATABASE>							        freshness_vector;
        vector<vector<DATABASE>::iterator>				    whom_vector;  // stores the list of friends to whom to send the message to
        vector<vector<DATABASE>::iterator>				    what_vector;  // stores the list of friends to whom to send the message to

        // for shuffling and selecting friends and what to send etc.
        int whom_array[MAX_FRIENDS];
        int what_array[MAX_FRIENDS];
        unsigned what_size;
        unsigned whom_size;
        unsigned heartbeat;

        // metrics
        unsigned long messages_sent;
        unsigned long bytes_sent;
        double average_freshness;
        double worst_freshness;
        int database_size;
        double oracle_freshness;
        double worst_oracle_freshness;
        unsigned worst_heartbeat;
        unsigned  worst_id;
        unsigned count_of_fresh_items;  // those above 90% say




        Peer(unsigned long id)
        {
            if(DEBUG>=2)
            {
                cout << "Entering constructor of Peer : " << id << endl;
            }
            myid = id;
            database_vector.reserve(MAX_FRIENDS);
            freshness_vector.reserve(MAX_FRIENDS);
            database_vector.clear();
            freshness_vector.clear();

            heartbeat=0;

            messages_sent=0;
            bytes_sent=0;
            database_size=0;
            worst_freshness=0;
            average_freshness=0;
            oracle_freshness=0;

            // add myself to the database list
            DATABASE d;
            d.ip=myid;
            d.latency=0;
            d.freshness=100;
            d.valid=0;
            database_add(d);

            update_myinfo();

            // add an send event for myself
            EventData event;
            event.time = global.current_time + 0.5;
            event.to = myid;
            event.type = SEND;
            event.from = myid;
            global.event_manager.add_event(event);

            if(DEBUG>=2)
                cout << "Peer: " << myid << " Scheduling a send event for myself at time " << event.time << endl;

        }


        void print_vectors()
        {
            vector<DATABASE>::iterator	p;

            cout << "Printing the latency vector....\n";
            /* check if friend already exists */
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                p->print();
            }
            cout << "Printing the freshness vector....\n";
            /* check if friend already exists */
            for(p = freshness_vector.begin(); p != freshness_vector.end(); p++)
            {
                p->print();
            }

        }



        void collect_metrics()
        {
            vector<DATABASE>::iterator p;
            double latest_freshness;

            average_freshness=0;
            worst_freshness=100;
            oracle_freshness=0;
            worst_oracle_freshness=1;
            database_size=0;
            count_of_fresh_items=0;
            worst_heartbeat=0;

            int count=0;

            int min_id;

            // for computing the average and worst freshness
            for(p = freshness_vector.begin(); p != freshness_vector.end(); p++)
            {
                if(p->valid==0)
                    continue;

                count++;
                // this entry could be old, check how old it is ....
                latest_freshness=global.decay_freshness(p->freshness,global.current_time-p->receipt_time);

                if(p->ip==myid)
                {
                    latest_freshness = 100;
                }

                average_freshness+=latest_freshness;
                if(latest_freshness<worst_freshness && p->valid==1)
                {
                    min_id = p->ip;
                    worst_freshness=latest_freshness;
                }
                if(p->valid==1)
                {
                    database_size++;
                }

                // now compute the real freshness of each database item
                p->oracle_freshness=global.oracle_freshness(global.peer[p->ip]->heartbeat-p->heartbeat);
		if(p->oracle_freshness < 100) {
			global.heart_beat_distribution[(int)p->oracle_freshness]++;
		}

		if(p->oracle_freshness<5)
                    count_of_fresh_items++;
                oracle_freshness+=p->oracle_freshness;
 
                /*
                if(global.peer[p->ip]->heartbeat-p->heartbeat>worst_heartbeat)
                {
                    //cout <<  global.peer[p->ip]->heartbeat << "," <<  p->heartbeat << endl;
                    worst_heartbeat=global.peer[p->ip]->heartbeat-p->heartbeat;

                }
                */
                if(p->oracle_freshness<worst_oracle_freshness)
                {
                    worst_oracle_freshness=p->oracle_freshness;
                }
           }

            average_freshness/=float(count);
            oracle_freshness/=float(count);

            if(DEBUG>=2)
                cout << "collect_metrics(): Myid = " << myid << " Average_freshness = " <<average_freshness  <<  " Oracle_freshness = " << oracle_freshness << endl;
        }

        void what(vector<DATABASE>::iterator whom) // name of friend to whom I am sending
        {
            vector<DATABASE>::iterator	  p;
            int i,size;

            what_vector.clear();
            switch(global.what_scheme)
            {
                case BROADCAST:
                    /* MULTICAST model , add all the people in the database */
                    for(p = freshness_vector.begin(); p != freshness_vector.end(); p++)
                    {
                        if(p->valid==1)
                            what_vector.push_back(p);
                    }
                    break;

                case RANDOM:
                    size=freshness_vector.size();
                    what_size=min(size,global.what);

                    for(i=0;i<size;i++)
                    {
                        what_array[i]=i;
                    }

                    // shuffle the array
                    for(i=0;i<what_size;i++)
                    {
                        int r=rand()%size;
                        int t=what_array[i];
                        what_array[i]=what_array[r];
                        what_array[r]=t;
                    }
                    for(i=0;i<what_size;i++)
                    {
                        what_vector.push_back(&freshness_vector[what_array[i]]);
                    }
                    break;

                case SPATIAL:
		    /* Simply send the top "what" pieces of information,
		     * instead of random
		     */
                    size=freshness_vector.size();
		    what_size=min(size,global.what);
		    
                    for(i=0;i<what_size;i++)
                    {
                        what_vector.push_back(&freshness_vector[i]);
                    }
                    break;

                default:
                    break;
            }

        }


        void whom(void)
        {
            vector<DATABASE>::iterator	  p;
            int i,size;

            whom_vector.clear();
            switch(global.whom_scheme)
            {

                case BROADCAST:
                    /* MULTICAST model , add all the people in the database */
                    for(p = database_vector.begin(); p != database_vector.end(); p++)
                    {
                        if(p->ip != myid )
                            whom_vector.push_back(p);
                    }
                    break;

                case RANDOM:
                    size=database_vector.size();
                    whom_size=min(size,global.whom);

                    for(i=0;i<size;i++)
                    {
                        whom_array[i]=i;
                    }

                    // shuffle the array
                    for(i=0;i<whom_size;i++)
                    {
                        int r=rand()%size;
                        int t=whom_array[i];
                        whom_array[i]=whom_array[r];
                        whom_array[r]=t;
                    }

                    for(i=0;i<whom_size;i++)
                    {
                        whom_vector.push_back(&database_vector[whom_array[i]]);
                    }
                    break;

                case SPATIAL:
		    
                    size=database_vector.size();
//		    if(size) 
//			    whom_vector.push_back(&database_vector[0]);

                    for(i=1;i<size;i++)
                    {
                        if(double(rand())/RAND_MAX <= pow(i, global.ro)) {
                            whom_vector.push_back(&database_vector[i]);
			}
                    }

                    break;

                default:

                    break;
            }

        }

        // inserts an empty entry into the database for a new ip address.
        // used just in the initializion phase of the program
        int add_friend(unsigned long ip)
        {
            /*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
            vector<DATABASE>::iterator	p;
            /*~~~~~~~~~~~~~~~~~~~~~~~~~~*/

            if(DEBUG >= 2)
                cout << "Entering add_friend()\n";

            /* check if friend already exists */
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                if(p->valid && p->ip == ip)
                    return -1;        // friend already exists
            }


            DATABASE	d;
            d.ip = ip;
            d.valid = 0;
            d.latency = global.get_latency(myid, ip);
            d.freshness = -1;
            database_add(d);

            if(DEBUG >= 2)
                cout << "Leaving the add_friend()\n";
            return 0;
        }

        /*
          ===================================================================================================================
             adds an element to the current database in a sorted order
          ===================================================================================================================
          */
        void database_add(DATABASE d)
        {
            if(DEBUG>=2)
            {
                cout << "\nentering database_add(), This is peer : " << myid << " Adding database entry : ";
                d.print();
                cout << endl;
            }

            /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
            vector<DATABASE>::iterator						p;

            /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

            // delete if already there

            for(p=database_vector.begin();p!=database_vector.end();p++)
            {
                if(p->ip==d.ip)
                {
                    /* ankit: first check if this is fresher information. dont
                    * change to old information!
                    */
                    double present_freshness=global.decay_freshness(p->freshness,global.current_time-p->receipt_time);
                    if(present_freshness > d.freshness)
                        return; /* get lost. we have newer information! */

                    database_vector.erase(p);
                    break;
                }
            }


            for(p=freshness_vector.begin();p!=freshness_vector.end();p++)
            {
                if(p->ip==d.ip)
                {
                    freshness_vector.erase(p);
                    break;
                }
            }


            /* add to latency vector */
            bool flag=false;
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                if(p->latency > d.latency)
                {
                    database_vector.insert(p, d);
                    flag=true;
                    break;
                }
            }

            if(flag==false) // if empty !
                database_vector.push_back(d);


            /* add to freshness vector */
            flag=false;
            for(p = freshness_vector.begin(); p != freshness_vector.end(); p++)
            {
                if(p->freshness < d.freshness)
                {
                    freshness_vector.insert(p, d);
                    flag=true;
                    break;
                }
            }

            if(flag==false) // if empty !
                freshness_vector.push_back(d);

        }


        DATABASE database_find(unsigned ip)
        {
            vector<DATABASE>::iterator	p;

            /* check if friend already exists */
            for(p = database_vector.begin(); p != database_vector.end(); p++)
            {
                if(p->ip==ip)
                {
                    return (*p);
                }
            }

        }

        void update_database(DATABASE d)
        {

            if(DEBUG>=2)
            {
                cout << "\nentering update_database(), This is peer : " << myid << " Updating database entry : ";
                d.print();
                cout << endl;
            }

            vector<DATABASE>::iterator						p;
            bool flag=false;

            /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

            // delete if already there

            for(p=database_vector.begin();p!=database_vector.end();p++)
            {
                if(p->ip==d.ip)
                {
                    /* ankit: first check if this is fresher information. dont
                    * change to old information!
                    */
                    double present_freshness=global.decay_freshness(p->freshness,global.current_time-p->receipt_time);
                    if(present_freshness > d.freshness)
                        return; /* get lost. we have newer information! */

                    *p=d;
                    flag=true;
                    break;
                }
            }
            if(!flag)
                database_vector.push_back(d);


            flag=false;
            for(p=freshness_vector.begin();p!=freshness_vector.end();p++)
            {
                if(p->ip==d.ip)
                {
                    *p=d;
                    flag=true;
                    break;
                }
            }

            if(!flag)
                freshness_vector.push_back(d);

        }

        // update the database with updated information
        // requires updating the latency_vector and freshness_vector also, which is tricky

        void update_myinfo()
        {
            //time_t t;
            timeb t;
            ftime(&t);

            heartbeat++;

            DATABASE d=database_vector[0];
            //sprintf(d.information.information,"This is me, Mr. %u,Time is %s, millis = %u",myid,ctime(&t.time),t.millitm);

            d.valid=1;
            d.freshness=100;
            //d.information.length=strlen(d.information.information)+1;
            d.information.length=10;
            d.information.ip=myid;
            d.receipt_time=global.current_time;
            d.heartbeat=heartbeat;
            database_add(d);

            // increase by heartbeat , used for computing "REAL" freshness by other peers



        }
        /*
         ===================================================================================================================
            schedule recieve information events for desired number of friends
         ===================================================================================================================
         */
        void event_send(EventData ev)
        {
            /*~~~~~~~~~~~~~~*/
            int			i, j,to;
            EventData	event;
            vector<DATABASE>::iterator what_iterator;
            /*~~~~~~~~~~~~~~*/

            if(DEBUG>=2)
                cout << "Entering event_send() , myid = " << myid << endl;

            whom();
            if(whom_vector.size()==0)
            {
                if(DEBUG>=2)
                    cout << "Peer: " << myid << ", My send vector is empty !" << endl;
            }
            else
            {
                if(DEBUG>=2)
                    cout << "Peer: "  << myid << " whom_vector size is " << whom_vector.size() << endl;
            }

            update_myinfo();

            for(i = 0; i < whom_vector.size(); i++)
            {
                to = whom_vector[i]->ip;
                if(to==myid)
                    continue; //never send to myself
                if(DEBUG>=2)
                    cout << "Sending information to peer:  " << to << endl;

                // what to send
                what(whom_vector[i]);

                for(j = 0; j < what_vector.size(); j++)
                {
                    if(what_vector[j]->valid==0)
                        continue;

                    /* schedule a recieve event */
                    event.time = global.current_time + global.get_latency(myid, to);
                    event.to = to;
                    event.type = RECEIVE;
                    event.from = myid;
                    event.heartbeat=what_vector[j]->heartbeat;

                    // update the freshness of the item which we are sending to others....
                    event.freshness = global.decay_freshness(what_vector[j]->freshness,global.current_time-what_vector[j]->receipt_time);
                    event.information = what_vector[j]->information;
                    event.information.ip=what_vector[j]->ip;    // to whom this information belongs

                    global.event_manager.add_event(event);

                    //update the metrics
                    messages_sent++;
                    bytes_sent+=sizeof(EventData)-sizeof(event.time)-sizeof(event.to)-sizeof(event.from)-sizeof(event.type);
                }

                if(DEBUG>=2)
                    cout << "Peer : " << myid << ", scheduled a recieve event for " << to << " at time " << event.time << endl;


            }

            /* schedule a send event for myself */

            // next send event between 3 and 10
            double next_send=20.0+(10.0*rand()/(RAND_MAX+1.0));

            event.time = global.current_time + next_send;
            event.to = myid;
            event.type = SEND;
            event.from = myid;
            //strcpy(event.information.information,"Don't Care!");

            global.event_manager.add_event(event);

            if(DEBUG>=2)
                cout << "Peer: " << myid << " Scheduling a send event for myself at time " << event.time << endl;

        }

        /*
         ===================================================================================================================
            got a packet from a peer !
         ===================================================================================================================
         */
        void event_recieve(EventData event)
        {
            /* just add this packet to the database, that's all; */

            if(DEBUG >= 2)
                cout << "Entering event_recieve, myid = " << myid << endl;

            /* update the freshness */
            event.freshness = global.decay_freshness(event.freshness, global.get_latency(event.from, event.to));

            /* ideally also look at originating timestamp */
            DATABASE d;

            d.ip=event.information.ip;
            d.latency=global.get_latency(myid,d.ip);
            d.freshness=event.freshness;
            d.information=event.information;
            d.heartbeat=event.heartbeat;
            d.valid=1;
            d.receipt_time=global.current_time;

            update_database(d);
            //update the database
            if(DEBUG>=2)
            {
                cout << "Peer : " << myid << " , Printing my database now...\n";
                print_vectors();
            }

            // till now we don't really call the add_friend function.
        }
};

// for 1 hr atleast
void insert_probe_events(double n)
{
    EventData event;
    double t;


    event.time=0;
    event.type=PROBE;
    event.from=0;
    event.to=0;
    //strcpy(event.information.information,"Probe !");

    for(t=n;t<=6000000;t+=n)
    {
        event.time=t;
        global.event_manager.add_event(event);
    }

    cerr << "\ninsert_probe_events(): Probes inserted at intervals of " << n << " seconds" << endl;

}


void main(int argc,char* argv[])
{
    int peers;
    unsigned event_step=0;
    EventData event;

    double sd_oracle_freshness=0;
    double sd_decay_freshness=0;
    double sd_fresh_items=0;

    ofstream f;

    if(argc<=2)
    {
        cout << "Usage : " << argv[0] << " topology_file scheme other_paramaters...\n";
        exit(0);
    }

    // first read the topology

    global.topology.read_topology(argv[1]);

    // set the number of peers
    global.no_of_peers=global.topology.v;
    peers=global.no_of_peers;



    // set the gossip scheme;
    //

    int argk;


    if(strcmp(argv[2],"broadcast")==0)
    {
    	global.gossip_scheme=BROADCAST;
    	cout << "# using scheme broadcast" << endl;
    }
    if(strcmp(argv[2],"random")==0)
    {
    	global.gossip_scheme=RANDOM;
    	if(argc!=5)
    	{
    		cout << "Please enter whom and what parameters" << endl;
    		exit(-1);
    	}
    	global.whom=atoi(argv[3]);
    	global.what=atoi(argv[4]);
    	cout << "# using scheme random. whom = " << global.whom << " , what = " << global.what << endl;

    }
    
    if(strcmp(argv[2],"spatial")==0)
    {
    	global.gossip_scheme=SPATIAL;
    	if(argc!=5)
    	{
    		cout << "\nPlease enter max_what and ro" << endl;
    		exit(-1);
    	}
    	global.what=atoi(argv[3]);
    	cout << "# using scheme spatial. max_what = " << global.what  << endl;

    	global.ro=atof(argv[4]);
    	cout << "# using scheme spatial. ro = " << global.ro  << endl;
    }


    /*
    global.gossip_scheme=SPATIAL;
    global.ro=1.4; // partiality constant or skew factor
    global.cx=5; //normalizing constant
    */

    /*
    global.gossip_scheme=BROADCAST;

    */

    // init the peers, set their contact points
    cerr << "Initializing the peers...\n";

    int i,j;
    for(i=0;i<peers;i++)
    {
        global.peer[i]=new Peer(i);
        if(i>0)
            global.peer[i]->add_friend(i-1);
    }


    // insert the probe events;
    insert_probe_events(20); // every n seconds , do the probe

    cerr << "Now starting the event_handler loop\n";
    // start the event handler from now on

    cout << "# GNUPLOT file containing the metrics for plotting...." << endl;
    cout << "# 1time 2messages_sent 3bytes_sent 4average_freshness 5worst_freshness 6oracle_freshness 7database_size 8no_of_peers" << endl;
    cout << "# 9count_of_fresh_items 10dev_of_count_of_fresh_items 11dev_of_decay_freshness 12deviation_in_oracle_freshness" << endl;
    cout << endl;

    while(1)
    {
        event=global.event_manager.get_min_event();
        event_step++;
        global.current_time=event.time;

        // cerr << "Event step : " << event_step << ", Current time : " << global.current_time << ", type = " << event.type <<  endl;

        if(DEBUG>=2)
            cout << "\n********Step : " << event_step << " , Extracted event at time " << global.current_time << endl;
        //event.print();

        switch(event.type)
        {
            case RECEIVE:
                global.peer[event.to]->event_recieve(event);
                break;

            case SEND:
                global.peer[event.to]->event_send(event);
                break;

            case PROBE:
                cerr << "****** Probe Event ! at time " << global.current_time;

                global.messages_sent=0;
                global.bytes_sent=0;
                global.average_freshness=0;
                global.average_oracle_freshness=0;
                global.worst_freshness=100;
                global.average_database_size=0;
                global.count_of_fresh_items=0;
		memset(global.heart_beat_distribution, 0, sizeof(unsigned) * 100);

                for(j=0;j<peers;j++)
                {
                    global.peer[j]->collect_metrics();
                    if(PERPEER)
                    {
                        cerr << "**********************************" << endl;
                        cerr << "Metrics for peer " << j << endl;
                        cerr << "Messages_sent = " << global.peer[j]->messages_sent << endl;
                        cerr << "Bytes_sent = " << global.peer[j]->bytes_sent << endl;
                        cerr << "Average_Freshness = " << global.peer[j]->average_freshness << endl;
                        cerr << "Worst_freshness = " << global.peer[j]->worst_freshness << endl;
                        cerr << "Oracle Freshness = " << global.peer[j]->oracle_freshness << endl; // higher , worse it is
                        cerr << "Fresh items = " << global.peer[j]->count_of_fresh_items << endl;

                        cerr << "Database size = " << global.peer[j]->database_size << "/" << global.no_of_peers << endl;
                    }

                    global.messages_sent+=global.peer[j]->messages_sent;
                    global.bytes_sent+=global.peer[j]->bytes_sent;
                    global.average_freshness+=global.peer[j]->average_freshness;
                    global.average_oracle_freshness+=global.peer[j]->oracle_freshness;
                    global.average_database_size+=global.peer[j]->database_size;
                    global.count_of_fresh_items+=global.peer[j]->count_of_fresh_items;

                    if(global.peer[j]->worst_freshness<global.worst_freshness)
                        global.worst_freshness=global.peer[j]->worst_freshness;
                }

                global.average_freshness/=float(peers);
                global.average_oracle_freshness/=float(peers);
                global.average_database_size/=float(peers);
                global.count_of_fresh_items/=float(peers);

                sd_oracle_freshness=0;
                sd_decay_freshness=0;
                sd_fresh_items=0;

		f.open("hb_dist");
		
		
		for(j=0; j<100; j++) {

			global.heart_beat_distribution[j]/=float(peers);
			f << j << "\t" << global.heart_beat_distribution[j] << endl;
		}
		f.close();
		

                // now compute deviation from the average
                for(j=0;j<peers;j++)
                {
                    sd_oracle_freshness+=pow((global.peer[j]->oracle_freshness-global.average_oracle_freshness),2);
                    sd_decay_freshness+=pow((global.peer[j]->average_freshness-global.average_freshness),2);
                    sd_fresh_items+=pow((global.peer[j]->count_of_fresh_items-global.count_of_fresh_items),2);
                }

                sd_oracle_freshness=sqrt(sd_oracle_freshness/float(peers));
                sd_decay_freshness=sqrt(sd_decay_freshness/float(peers));
                sd_fresh_items=sqrt(sd_decay_freshness/float(peers));

                cerr << "\n*******************************" << endl;
                cerr << "GLOBAL METRICS " << endl;
                cerr << "**********************************" << endl;
                cerr << "Gossip scheme : " << s_SCHEMES[global.gossip_scheme] << endl;
                cerr << "Current time : " << global.current_time << endl;
                cerr << "Messages_sent = " << global.messages_sent << endl;
                cerr << "Bytes_sent = " << global.bytes_sent << endl;
                cerr << "Average_Freshness = " << global.average_freshness << endl;
                cerr << "Worst_freshness = " << global.worst_freshness << endl;
                cerr << "Average Oracle_freshness = " << global.average_oracle_freshness << endl;
                cerr << "Average Database size = " << global.average_database_size << "/" << global.no_of_peers << endl;

                // gnuplot metrics
                cout << setw(6) << global.current_time << setw(12) <<  global.messages_sent << setw(12) << global.bytes_sent << setw(12) << global.average_freshness
                << setw(12) <<  global.worst_freshness << setw(12) <<  global.average_oracle_freshness << setw(12) << global.average_database_size << setw(12) << global.no_of_peers;
                cout << setw(12) << global.count_of_fresh_items << setw(12) << sd_fresh_items << setw(12) << sd_decay_freshness << setw(12) << sd_oracle_freshness;
                cout << endl;


                break;
            default:
                break;
        }
    }
}
